#pragma rtGlobals=3		// Use modern global access method.
#pragma TextEncoding 	= "UTF-8"
#pragma IgorVersion		= 8.0
#pragma ModuleName		= SpectraWatchDog
#pragma version			= 1.10
#pragma DefaultTab		= {3,20,4}			// Set default tab width in Igor Pro 9 and later

static Constant kVersion = 1.10
static StrConstant kVersionDate = "04/2025"

// --------------------- Project Updater header ----------------------
// If you're using Igor Pro 8 or later and have Project Updater
// installed, this package can check periodically for new releases.
// https://www.wavemetrics.com/project/Updater
static Constant kProjectID = 22182																// the project node on IgorExchange
static StrConstant ksShortTitle = "Spectra WatchDog"											// the project short title on IgorExchange

//________________________________________________________________________________________________
//	Written by Stephan Thuermer - https://www.wavemetrics.com/user/chozo
//	
//	Spectra Folder Watchdog
//	
//	Writes text logs of spectra metadata from files in a designated folder. Logs can be complemented
//	with user information, i.e., the log is updated with values entered in a panel. Folders can either
//	be read once or watched continuously for new spectra files.
//	
//	Provided with support for the following file formats:
//		- ibw:	Scienta SES photoemission spectra
//		- zip:	Zip files containing spectra (default content: *.ibw)
//
// 2023-06-20 - ver. 1.01:	Added support for zipped IBW files (Igor 9).
// 2023-07-28 - ver. 1.02:	Initial public release.
//							Fixed bug: Entries were not updated correctly for duplicate file names within zip files.
//							Added support for extension lists.
//							Generalized info function support.
// 2023-11-24 - ver. 1.03:	Fixed bug: Processing did not work when multiple parser functions for the same file type were present.
//							A message is printed if a file is skipped due to a failed parsing.
// 2025-04-22 - ver. 1.10:	File entries are now sorted in case of multiple new ones.
//							Table properties can now be defined on a per-extension basis via a "WDinfo_EXT_NAME()" function.
//							Made the settings file optional. Use kUseSettingsFile = 0 to only save parameters in the current experiment.
//							The watchdog panel now uses the available screen space.
//							Resetting the WD state optionally sets a new log file name.
//							Added a more accurate time calculation into the SES data parser.
//							Table column widths are now set correctly, i.e., in fine increments.
//							Color numbers are now indicated in the list box right-click menu.
//							The SES parser now reads wave info without actually loading the data.
//
// Possible future features:
//	- list-box table sorting
//	- reading log files via column headers, i.e., support for files in arbitrary order
//________________________________________________________________________________________________

// *** global definitions ************************************************************************
static Constant kCheckFolderPeriod		= 60													// In "ticks" which are ~1/60th of a second.
static Constant kUseSettingsFile		= 1														// whether or not to use external save file
static StrConstant kLogFileEnding		= ".log"												// writes text files with this ending
static StrConstant kDelimiterChar		= "\t"													// type of delimiter for entries in the log file
static StrConstant kSettingsFileName	= "Spectra WatchDog settings.dat"
static StrConstant kWatchdogWorkingDir 	= "Spectra_Watchdog"
static StrConstant kDefaultExtensionStr	= "ibw"													// looks for this file format

Menu "Spectra Tools"
	"-"
	"Folder Watchdog...",	/Q, SpectraWatchDog#buildWatchDogPanel()
	"Folder Quick-log...",	/Q, SpectraWatchDog#QuickLogFromFolder()
End

// *** parser function definitions *** ###########################################################

function/S WDinfo_ZIP_xSESx(variable request)													// zip:	Scienta SES photoemission spectra
	return WDinfo_IBW_xSESx(request)
end
function/S WDinfo_IBW_xSESx(variable request)													// ibw:	Scienta SES photoemission spectra
	string info = ""
	Switch (request)
		case WDInfo_DefaultColumnWidths:
			info = "2;2.5;1;1;1;1;1;1;1;1;1;1;2.5;2.5;1.5;1.25;1.25;"
		break
		case WDInfo_DefaultColumnColors:
			info = "0;1;0;0;0;0;0;0;0;1;4;4;8;8;0;0;3;"
		break
		case WDInfo_DefaultRepeatColumn:
			info = "0;0;0;0;0;0;0;0;0;0;0;0;0;0;0;0;0;"
		break
	EndSwitch
	return info
end

function/wave WDfunc_IBW_xSESx(string filePath)													// ibw:	Scienta SES photoemission spectra
	string infoStr = readIBWnotes(filePath)
	wave/Z waveDim = readIBWdimensions(filePath)
	if (!strlen(infoStr) || !StringMatch(infoStr[0,4],"[SES]"))									// look in the notes for the SES header
		return $""
	endif
	// ++++++++++++++++++++++++++++++++++++++++
	string LoadPath  = GetStr("LoadPath", infoStr)
	string FullPath  = GetStr("File", infoStr)
	string SweepNum  = GetStr("Number of Sweeps", infoStr)
	string StartDate = GetStr("Date", infoStr)
	string StartTime = GetStr("Time", infoStr)
	
	int yDim = waveDim[1], zDim = waveDim[2]
	string TagName = ParseFilePath(3, FullPath, SelectString(strsearch(FullPath,"\\",0) > -1,":","\\"), 0, 0)
	if (!strlen(TagName))
		TagName = ParseFilePath(3, filePath, SelectString(strsearch(filePath,"\\",0) > -1,":","\\"), 0, 0)
		TagName = RemoveEnding(TagName,GetStr("Spectrum Name", infoStr))
	endif
	if (StringMatch(GetStr("Name", infoStr), "Add Dimension"))
		SweepNum = SelectString(str2num(GetStr("Number of Slices", infoStr)) > 1, num2str(yDim), num2str(zDim == 0 ? 1 : zDim))
	endif
	
	variable tStep = GetVal("Step Time",infoStr)/1000											// the end time is calculated as an offset to the start
	variable eStep = GetVal("Energy Step",infoStr)
	variable ePass = GetVal("Pass Energy",infoStr)
	variable scanT = ((GetE(infoStr,1)-GetE(infoStr,0) + 0.0883*ePass)/eStep * (tStep + 0.0015) + 0.55) * str2num(SweepNum)
	
	// ++++++++++++++++++++++++++++++++++++++++
	Make/Free/T/N=(0,2) info
	info[DimSize(info,0)] = {{"Tag"}			,{TagName}}
	info[DimSize(info,0)] = {{"Sample"}			,{GetStr("Sample", infoStr)}}
	info[DimSize(info,0)] = {{"E_Pass [eV]"}	,{GetStr("Pass Energy", infoStr)}}
	info[DimSize(info,0)] = {{"E_Mode"}			,{SelectString(StringMatch(GetStr("Energy Unit", infoStr), "Kinetic"), "BE", "KE")}}
	info[DimSize(info,0)] = {{"E_Start [eV]"}	,{GetStr("Low Energy", infoStr)}}
	info[DimSize(info,0)] = {{"E_End [eV]"}		,{GetStr("High Energy", infoStr)}}
	info[DimSize(info,0)] = {{"E_Step [eV]"}	,{GetStr("Energy Step", infoStr)}}
	info[DimSize(info,0)] = {{"Dwell Time [ms]"},{GetStr("Step Time", infoStr)}}
	info[DimSize(info,0)] = {{"# Sweeps"}		,{SweepNum}}
	info[DimSize(info,0)] = {{"E_hv [eV]"}		,{GetStr("Excitation Energy", infoStr)}}
	info[DimSize(info,0)] = {{"Det. Slit"}		,{""}}
	info[DimSize(info,0)] = {{"Bias [V]"}		,{""}}
	info[DimSize(info,0)] = {{"Purpose"}		,{""}}
	info[DimSize(info,0)] = {{"Comments"}		,{GetStr("Comments", infoStr)}}
	info[DimSize(info,0)] = {{"Start Date"}		,{StartDate}}
	info[DimSize(info,0)] = {{"Start Time"}		,{StartTime}}
	info[DimSize(info,0)] = {{"Scan Time"}		,{Secs2Time(scanT,3)}}
	return info
end

static function/S GetStr(string which, string infoStr)
	return StringByKey(which, infoStr, "=", "\r")
end
static function GetVal(string key, string notes)
	return NumberByKey(key, notes, "=", "\r")
end
static function GetE(string notes, int which)
	if (cmpstr(GetStr("Energy Scale",notes), "Binding") == 0)
		return (GetVal("Excitation Energy",notes) - GetVal(SelectString(which,"High","Low")+" Energy",notes))
	else
		return (GetVal(SelectString(1-which,"High","Low")+" Energy",notes))
	endif
end

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

static function/S readIBWnotes(string path)														// reads wave notes from .ibw file
	int fileID, version, chckSum = 0, wSize = 0, nSize = 0, fSize = 0
	Open/Z/R fileID as path
	if (V_flag != 0)
		return ""
	endif
	FBinRead/B=3/F=2/U fileID, version
	if (version < 5)
		Close fileID
		return ""
	endif
	
	FBinRead/B=3/F=2/U fileID, chckSum															// chckSum is only read to move position forward
	FBinRead/B=3/F=3/U fileID, wSize
	FBinRead/B=3/F=3/U fileID, fSize
	FBinRead/B=3/F=3/U fileID, nSize
	string notes = PadString("", nSize, char2num(" "))
	FSetPos  fileID, (wSize + fSize + 64)
	FBinRead fileID, notes
	Close fileID
	return notes
end

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

static function/Wave readIBWdimensions(string path)												// reads dimension size from .ibw file into free wave
	int fileID, version, points
	Open/Z/R fileID as path
	if (V_flag != 0)
		return $""
	endif
	FBinRead/B=3/F=2/U fileID, version
	if (version < 5)
		Close fileID
		return $""
	endif
	
	Make/D/Free/N=(4) dims
	FSetPos fileID, (68 + 64)
	FBinRead/B=3/F=3/U fileID, dims
	Close fileID
	return dims
end

// *** folder quick logging *** ##################################################################

function QuickLogFromFolder()																	// saves a log of all data files in the same folder as the data
	CtrlNamedBackground SpectraWatchDogTask, status
	if (NumberByKey("RUN", S_info))
		printMsg("Caution: Cannot write log while WatchDog is running.")
	endif
	
	int i, fileID
	string currExt, extList = "", allFuncs = getFuncList("", "WDfunc_", 8, 8192)
	for (i=0; i<ItemsInList(allFuncs); i++)
		SplitString/E=("WDfunc_([[:alpha:]]+)_") StringFromList(i,allFuncs), currExt
		if (FindListItem("*."+currExt, extList, ",") < 0)
			extList += "*." + LowerStr(currExt) + ","
		endif
	endfor
	extList = RemoveEnding(extList,",") + SelectString(IgorVersion()>=9,"",",*.zip")			// zip support only from Igor 9
	
	Open/D/R/F=("Data Files ("+extList+"):"+ReplaceString("*",extList,"")+";")/M="Select any data file from folder to write the log file." fileID
	if (strlen(S_Filename) > 0)
		string loadDF  = ParseFilePath(1, S_Filename, ":", 1, 0)
		string logFile = loadDF + ParseFilePath(0, S_Filename, ":", 1, 1) + kLogFileEnding
		NewPath/O/Q/Z WD_WatchFolderPath, loadDF
		DeleteFile/Z logFile
		string loadList = IndexedFile(WD_WatchFolderPath,-1,"."+ParseFilePath(4, S_Filename, ":", 0, 0))
		for (i=0; i<ItemsInList(loadList); i+=1)
			SpectraWatchDog#WD_processFile(loadDF+StringFromList(i, loadList), logFile=logFile)
		endfor
		KillPath/Z WD_WatchFolderPath
	endif
end

//################################################################################################
// *** WD core functions ***
//################################################################################################

Constant WDInfo_DefaultColumnWidths = 0
Constant WDInfo_DefaultColumnColors = 1
Constant WDInfo_DefaultRepeatColumn = 2

static function/S WD_getListDefaults(string requestedExt, int request)
	int i
	string info = "", funcList = getFuncList(requestedExt, "WDinfo_", 4, 4)
	for (i=0; i < ItemsInList(funcList); i++)
		FUNCREF WDdummy_InfoFunc infoFunc=$StringFromList(i,funcList)
		info = infoFunc(request)
		if (strlen(info))																		// valid output
			break
		endif
	endfor
	if (!strlen(info))
		info = SelectString(request == WDInfo_DefaultColumnWidths, "0", "1")					// set default values
	endif
	return RemoveEnding(ReplaceString(";",info,","), ",")
end

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

static function/S WD_checkColWidths(string list)
	Wave/T w_txt = ListToTextWave(list, ",")
	Make/Free/N=(DimSize(w_txt,0)) w_num = str2num(w_txt[p]);	variable div = sum(w_num)
	w_num = 720*w_num[p]/div;	wfprintf list, "%d,", w_num										// minimal width of ListBox: 765-25-20
	return RemoveEnding(list,",")
end

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

static function/wave WD_readSpectraInfo(string filePath)										// returns a two-column text wave with all parameters from the input file
	int i
	string funcList = getFuncList(ParseFilePath(4,filePath,":",0,0), "WDfunc_", 8, 8192)
	wave/T/Z info = $""																			// initialize wave reference
	for (i=0; i < ItemsInList(funcList); i++)
		FUNCREF WDdummy_ParseFunc infoFunc=$StringFromList(i,funcList)
		wave/T/Z info = infoFunc(filePath)
		if (WaveExists(info))																	// valid output
			break
		endif
	endfor
	return info																					// format: col[0] = table column titles, col[1] = table entries
end

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

function/wave WDdummy_ParseFunc(string filePath)
	return $""
end
function/S    WDdummy_InfoFunc(variable request)
	return ""
end

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

static function/S getFuncList(string requestedExt, string prefix, int valType, int paraType)
	string allFuncs = SortList(FunctionList(prefix+"*",";","KIND:2,VALTYPE:"+num2str(valType)+",PARAM_0_TYPE:"+num2str(paraType)), ";", 4)
	if (!strlen(requestedExt))
		return allFuncs
	endif
	int i; string list = "", currExt
	for (i=0; i < ItemsInList(allFuncs); i++)
		SplitString/E=(prefix+"([[:alpha:]]+)_") StringFromList(i, allFuncs), currExt
		if (!CmpStr(currExt,requestedExt))
			list += StringFromList(i, allFuncs) + ";"
		endif
	endfor
	return list
end

static function printMsg(string msg)
	Print "[WD v" + num2str(kVersion) + "](" + time() + ") " + msg
end

// *** main processing function *** ###############################################################

static function WD_processFile(string filePath [,string logFile, string extList])
	int i, error = 0, panelCall = ParamIsDefault(logFile)
	string folder = "", zipNamePrefix = "", fileList = filePath									// is the full path already
	string fileExt = ParseFilePath(4,filePath,":",0,0)
	
	if (!Cmpstr(fileExt,"zip"))																	// added zip support
		folder = SpecialDirPath("Temporary",0,0,0)
		error  = 1
		#if(IgorVersion() >= 9)																	// unzip is supported from Igor 9
			unzipFile/O/Z filePath, RemoveEnding(folder,":")
			error = V_flag
		#endif
		if (error)
			printMsg("Error: Could not unzip file. Zipped files not supported.")
			return -3
		endif
		
		string zip_ext = kDefaultExtensionStr
		if (!ParamIsDefault(extList))
			extList = RemoveFromList("zip",extList)
			zip_ext = SelectString(strlen(extList),zip_ext,extList)
		endif
		
		NewPath/O/Q tmpPath, folder; fileList = ""
		for (i=0; i<ItemsInList(zip_ext); i++)													// grab valid files from extraction location
			fileList += IndexedFile(tmpPath, -1, "."+StringFromList(i,zip_ext))
		endfor
		if (!ItemsInList(fileList))
			printMsg("Caution: Zipped file \"" + fileInPath(filePath)+ "\" did not contain any spectra. Skipping ...")
			return -4
		endif
		
		sPrintf zipNamePrefix, "[%s] ", fileInPath(filePath)
		KillPath/Z tmpPath
	endif
	
	// +++++++++++++++++++++++++++++++++++ parse file list +++++++++++++++++++++++++++++++++++++++
	for (i=0; i<ItemsInList(fileList); i++)
		filePath = folder + StringFromList(i,fileList)
		wave/Z/T logInfo = SpectraWatchDog#WD_readSpectraInfo(filePath)							// read info from the file => spits out a two-column text wave
		
		if (strlen(zipNamePrefix))																// cleanup
			DeleteFile/Z filePath
		endif
		
		if (!WaveExists(logInfo))																// not a supported file => abort
			printMsg("Skipping '"+fileInPath(filePath)+"': No valid info found.")
			error = -1
			continue
		endif
		
		InsertPoints 0,1, logInfo;
		logInfo[0][] = {{"File Name"},{fileInPath(filePath)}}									// the first entry is always the file name in the log
		
		if (panelCall)
			logFile = getLogPath()
		endif
		// +++++++++++++++++++++++++++++++ write the log file ++++++++++++++++++++++++++++++++++++
		string header = "", entries = ""
//		string entryID = "";	sPrintf entryID, "[%.4d] ", countLogFileEntries(logFile)
		logInfo[0][1] = zipNamePrefix + logInfo[0][1]											// add unique identifier to file name (in case of overlapping names)
		Duplicate/Free/T/RMD=[][0] logInfo, tempHeader
		Duplicate/Free/T/RMD=[][1] logInfo, tempEntries
		
		int logID = 0
		Open/Z logID as logFile
		if (V_Flag == 0)
			FStatus logID
			if (V_logEOF == 0)																	// is new file => add header	
				wfprintf header, "%s"+kDelimiterChar, tempHeader
				header = RemoveEnding(header, kDelimiterChar)+"\r\n"
			endif
			wfprintf entries, "%s"+kDelimiterChar, tempEntries
			FSetPos logID, V_logEOF
			FPrintF logID, header + RemoveEnding(entries, kDelimiterChar)+"\r\n"
			Close logID
		else
			printMsg("Error: The log file is locked by another program and cannot be updated. Make sure the log file is not open at the same time, e.g., in Excel.")
			error = -2
			break
		endif
		
		if (panelCall)																			// only for the watchdog panel => write the local log wave
			SpectraWatchDog#WD_addLogListEntry(logInfo)
		endif
	endfor
	return error
end

static function/S fileInPath(string path)
	return ParseFilePath(0,path,":",1,0)
end

//################################################################################################

static function WD_addLogListEntry(wave/T logInfo)
	DFREF stg = root:Packages:$(kWatchdogWorkingDir)
	SVAR listNames = stg:WD_logInfoFileNames
	SVAR colRepeat = stg:WD_RepeatColumn
	wave/T logList = stg:WD_logInfoDispList
	wave   selList = stg:WD_logInfoSelList
	
	int i, selSize = DimSize(selList,1), infoSize = DimSize(logInfo,0)-1
	if (DimSize(logList,1) < infoSize)
		Redimension/N=(-1, infoSize, -1) logList, selList
	endif
	
	if (strlen(logList[0][0]))																	// if first row already contains an entry
		Redimension/N=(DimSize(logList,0)+1,-1,-1) logList, selList
		for (i = 0; i < infoSize; i++)															// repeat entries if needed
			variable isRepeated = str2num(StringFromList(i,colRepeat,","))
			isRepeated = numtype(isRepeated) != 0 ? 0 : isRepeated
			if (isRepeated)
				logInfo[i+1][1] = logList[DimSize(logList,0)-2][i]								// use values from previous entry
			endif
		endfor
	else
		for (i = 0; i < infoSize; i++)															// write labels for the first time
			string title = cleanStr(logInfo[i+1][0],"\"|'|,|;|")[0,255]
			SetDimLabel 1,i,$title, logList
		endfor
	endif
	
	logList[DimSize(logList,0)-1][0,infoSize-1] = logInfo[q+1][1]
	selList[DimSize(selList,0)-1][][0] = 0x02
	if (DimSize(selList,1) > selSize)															// new column(s) added => make select-able
		selList[][selSize,][0] = 0x02
	endif
	SpectraWatchDog#WD_colorListWave(selList)
	listNames += logInfo[0][1] + ";"
	return 0
end

static function/S cleanStr(string str, string list [, string insert])
	int i; string replace = SelectString(ParamIsDefault(insert),insert,"")
	for (i = 0; i < ItemsInList(list,"|"); i++)
		str = ReplaceString(StringFromList(i,list,"|"), str, replace)
	endfor
	return str
end

//################################################################################################

//static function countLogFileEntries(string logFilePath)
//	int i=0, fileID
//	string read
//	Open/Z/R fileID as logFilePath
//	if (!V_flag)
//		do
//			FReadLine fileID, read; i++
//		while (strlen(read) > 0)
//		Close fileID
//	endif
//	return i > 0 ? i-1 : 1
//end

// *** UI background task *** ####################################################################

function WD_SpectraWatchFolderBT(STRUCT WMBackgroundStruct &s)
	DFREF stg = root:Packages:$(kWatchdogWorkingDir)
	NVAR startCnt = stg:WD_startCount
	NVAR runCount = stg:WD_runCounter
	runCount = (s.curRunTicks-startCnt)/60														// record passed time since start
	SpectraWatchDog#WD_checkNewFiles()
	return 0
end

//################################################################################################

static function WD_checkNewFiles()
	DFREF stg = root:Packages:$(kWatchdogWorkingDir)
	SVAR targetDir = stg:WD_TargetFolder
	SVAR doneList  = stg:WD_ProcessedFileList
	SVAR targetExt = stg:WD_TargeFileExt
	
	NewPath/O/Q WD_WatchFolderPath, targetDir													// make sure the path exists and is not overwritten elsewhere
	int i; string currFileList = ""
	for (i = 0; i < ItemsInList(targetExt); i++)
		currFileList += IndexedFile(WD_WatchFolderPath, -1, "."+StringFromList(i,targetExt))	// list of files in the folder to watch
	endfor
	string newFileList = SortList(RemoveFromList(doneList, currFileList), ";", 16)				// compare to already processed list
	
	for(i = 0; i < ItemsInList(newFileList); i++)
		string fileName = StringFromList(i, newFileList)
		if (SpectraWatchDog#FileIsNotLocked(targetDir+fileName,0))								// ensures that the file is closed before processing it to avoid processing while the other program is still writing it.
			SpectraWatchDog#WD_processFile(ParseFilePath(5,targetDir,":",0,0)+fileName, extList=targetExt)		// pass the file path as Mac formatted
			doneList += fileName+";"
		endif
	endfor
	return 0
end

//################################################################################################

static function FileIsNotLocked(string fullPath, variable doWrite)
	variable refNum = 0
	if (doWrite)
		Open/Z refNum as fullPath
	else
		Open/R/Z refNum as fullPath
	endif
	if (refNum != 0)																			// file successfully opened?
		Close refNum
		return 1
	endif
	return 0
end

//################################################################################################
// *** UI for WD ***
//################################################################################################

static function buildWatchDogPanel()
	if (WinType("SpectraWatchDogPanel") == 7)
		DoWindow/F SpectraWatchDogPanel
		return 0
	endif
	
	SpectraWatchDog#WD_prepareWorkingFolder()
	DFREF stg = root:Packages:$(kWatchdogWorkingDir)
	SVar colWidths = stg:WD_ColumnWidths
	
	variable TFTResX, TFTResY
	if (cmpstr(IgorInfo(2),"Macintosh")==0)		// Mac
		string TFTInfo = StringByKey("RECT",ReplaceString(",RECT",StringByKey("SCREEN1",IgorInfo(0),":"),";RECT"),"=")
		TFTResX = (Str2Num(StringFromList(2,TFTInfo,",")) - Str2Num(StringFromList(0,TFTInfo,",")))*(72/ScreenResolution)  
		TFTResY = (Str2Num(StringFromList(3,TFTInfo,",")) - Str2Num(StringFromList(1,TFTInfo,",")))*(72/ScreenResolution)  
	else										// Windows
		GetWindow kwFrameInner wsizeDC
		TFTResX = (v_right-v_left)
		TFTResY = (v_bottom-v_top)
	endif
	variable wT = 70, wL = 150, mW = 765, mH = 300
	variable wB	= TFTResY - 30; wB = limit(wB, wT+mH, inf)
	variable wR	= TFTResX - wL; wR = limit(wR, wL+mW, inf)
	
	NewPanel/K=1/N=SpectraWatchDogPanel/W=(wL,wT,wR,wB) as "Spectra Watchdog (ver. "+num2str(kVersion)+" - "+kVersionDate+")"
	variable px = PanelResolution("SpectraWatchDogPanel")/ScreenResolution
	SetWindow SpectraWatchDogPanel, sizelimit={mW*px, mH*px, inf, inf}
	
	Setvariable v_TargetFolder	,pos={11,8}		,size={mW-240,23}	,bodywidth=mW-340	,title="Folder to Watch:"		,help={"Sets the folder that the Watchdog watches."}
	Setvariable v_LogFileFolder	,pos={11,34}	,size={mW-240,23}	,bodywidth=mW-340	,title="Save Log File to:"		,help={"Sets the folder to which the output log is written to."}
	Button b_SelectTargetFolder	,pos={mW-223,7}	,size={70,20}							,title="Select"
	Button b_SelectOutLogFolder	,pos={mW-223,33},size={70,20}							,title="Select"
	Button b_ClearLogList		,pos={mW-143,60},size={130,20}							,title="Clear Display List"		,help={"Clears the currently displayed list (NOT the log file). Already processed files will not be reprocessed."}
	Button b_ResetWatchDog		,pos={mW-253,60},size={100,20}							,title="Reset All"				,help={"Completely resets watchdog to the initial state. All files will reprocessed after restart."}
	Button b_LoadFromLogFile	,pos={mW-143,33},size={130,20}							,title="Read Log Entries"		,help={"Loads past log file entries into the current watchdog session."}
	Button b_StartStopButton	,pos={180,60}	,size={140,20}							,title="Start Watchdog"			,help={"Starts or stops the Watchdog task."}
	Button b_ReadOnceButton		,pos={330,60}	,size={140,20}							,title="Read Folder Once"		,help={"Just check the folder once without starting the background task."}
	Setvariable v_TargetFileExt	,pos={mW-135,8}	,size={123,23}		,bodywidth=55		,title="File Type(s): *."		,help={"Sets the file type (ending) which the watchdog will look out for. Only supported file types should be entered here. Multiple extensions can be watched at the same time. For this, enter a semicolon-separated list."}
	CheckBox c_ReadOnlyNew		,pos={mW,9}		,size={140,23}							,title="Ignore Existing Files"	,help={"Ignores already present files in the selected folder when starting the watchdog. Only newly added files will be processed."}
	Setvariable v_runCountDisp	,pos={11,61}	,size={160,23}		,bodywidth=60		,title="Running for:"

	ListBox WD_DataDisplayList	,pos={11,85}	,size={wR-wL-22,wB-wT-96}
	ListBox WD_DataDisplayList	,listWave = stg:WD_logInfoDispList
	ListBox WD_DataDisplayList	,selWave  = stg:WD_logInfoSelList
	ListBox WD_DataDisplayList	,colorWave= stg:WD_logInfoColors
	ListBox WD_DataDisplayList	,proc=WD_ListHook ,frame=1 ,focusRing=0 ,mode=10 ,editStyle=1 ,userColumnResize=1 ,clickEventModifiers=5
	Execute/Q/Z "ListBox WD_DataDisplayList ,win=SpectraWatchDogPanel ,widths={"+colWidths+"}"
	
	Button b_SelectTargetFolder	,proc=WD_SelectInputFolderProc
	Button b_SelectOutLogFolder	,proc=WD_SelectLogFileFolderProc
	Button b_ClearLogList		,proc=WD_ClearAndResetButton
	Button b_ResetWatchDog		,proc=WD_ClearAndResetButton		,fstyle=2
	Button b_LoadFromLogFile	,proc=WD_ReadFromLogButton
	Button b_StartStopButton	,proc=WD_StartTaskButton			,fstyle=1	,valueColor=(0,16000,65000)
	Button b_ReadOnceButton		,proc=WD_StartTaskButton			,fstyle=1	,valueColor=(0,40000,16000)
	Setvariable v_TargetFolder	,value=stg:WD_TargetFolder	,proc=WD_SetLogFileFolderProc
	Setvariable v_TargetFileExt	,value=stg:WD_TargeFileExt	,proc=WD_SetLogFileFolderProc
	Setvariable v_LogFileFolder	,value=stg:WD_LogFilePath	,proc=WD_SetLogFileFolderProc
	Setvariable v_runCountDisp	,value=stg:WD_runCounter	,noedit=1	,limits={-inf,inf,0} ,format="%d s"
	
	SetWindow SpectraWatchDogPanel, hook(panelHook) = WD_PanelHook
	SpectraWatchDog#enableDisableControls()
	return 0
end

//################################################################################################

static function/S getLogPath()
	DFREF stg = root:Packages:$(kWatchdogWorkingDir)
	SVAR logPath = stg:WD_LogFilePath
	return logPath
end

//################################################################################################

static function enableDisableControls()															// disables all controls other than start button and runtime counter
	string ctList = RemoveFromList("b_StartStopButton;v_runCountDisp;WD_DataDisplayList;",ControlNameList("SpectraWatchDogPanel",";","*"))
	CtrlNamedBackground SpectraWatchDogTask, status;	int isRunning = NumberByKey("RUN", S_info)
	ModifyControlList ctList ,win=SpectraWatchDogPanel ,disable=2*isRunning
	Button b_StartStopButton ,win=SpectraWatchDogPanel ,title=SelectString(isRunning,"Start Watchdog","Stop Watchdog")
end

//################################################################################################

static function WD_prepareWorkingFolder()
	DFREF savDF = GetDataFolderDFR()
	string pkDF = "root:Packages:"+kWatchdogWorkingDir+":"
	NewDataFolder/O/S root:Packages
	NewDataFolder/O/S root:Packages:$(kWatchdogWorkingDir)
		string targetExt = StrVarOrDefault(pkDF+"WD_TargeFileExt", "")
		string colRepeat = StrVarOrDefault(pkDF+"WD_RepeatColumn", "")
		string colWidths = StrVarOrDefault(pkDF+"WD_ColumnWidths", "")
		string colorList = StrVarOrDefault(pkDF+"WD_ColumnColors", "")
		string targetDF  = StrVarOrDefault(pkDF+"WD_TargetFolder", "")
		string logFileDF = StrVarOrDefault(pkDF+"WD_LogFilePath", "")
		string doneFList = StrVarOrDefault(pkDF+"WD_ProcessedFileList", "")
		variable runNo   = NumVarOrDefault(pkDF+"WD_runCounter", 0)
		variable startNo = NumVarOrDefault(pkDF+"WD_startCount", 0)
				
		if (kUseSettingsFile)
			string settings = SpectraWatchDog#WD_readSettingsFromFile()							// load previous settings from disc
			targetExt = StringByKey("Spectra_Extension",settings, "=")
			colRepeat = StringByKey("Column_Repeat",	settings, "=")
			colWidths = StringByKey("Column_Widths",	settings, "=")
			colorList = StringByKey("Column_Colors",	settings, "=")
			targetDF  = StringByKey("Target", 			settings, "=")
			logFileDF = StringByKey("LogDir",			settings, "=")
		endif
		if (!strlen(targetExt))
			targetExt = kDefaultExtensionStr
		endif
		if (!strlen(colRepeat))
			colRepeat = WD_getListDefaults(targetExt, WDInfo_DefaultRepeatColumn)
		endif
		if (!strlen(colWidths))
			colWidths = WD_getListDefaults(targetExt, WDInfo_DefaultColumnWidths)
			colWidths = WD_checkColWidths(colWidths)
		endif
		if (!strlen(colorList))
			colorList = WD_getListDefaults(targetExt, WDInfo_DefaultColumnColors)
		endif
		
		String/G WD_TargetFolder = targetDF
		String/G WD_LogFilePath  = logFileDF
		String/G WD_TargeFileExt = targetExt
		String/G WD_RepeatColumn = colRepeat
		String/G WD_ColumnColors = colorList
		String/G WD_ColumnWidths = colWidths
		Variable/G WD_runCounter = runNo
		Variable/G WD_startCount = startNo
		String/G WD_ProcessedFileList = doneFList												// the list of files which have been read (logged or not)
		
		wave/Z/T list = WD_logInfoDispList														// the local log list
		if (!WaveExists(list))
			String/G WD_logInfoFileNames = ""													// NEEDS TO MATCH WD_logInfoDispList in size at all times => list of files which actually have been logged locally
			Make/T/N=(1,1) root:Packages:$(kWatchdogWorkingDir):WD_logInfoDispList/wave=list
		endif
		
		Make/O/N=(DimSize(list,0),DimSize(list,1),3) WD_logInfoSelList = 0
		SetDimLabel 2,1,backColors,WD_logInfoSelList											// define plane 1 as background colors
		SetDimLabel 2,2,foreColors,WD_logInfoSelList											// define plane 2 as foreground colors
		if (DimSize(list,1) > 1)
			WD_logInfoSelList[][][0] = 0x02														// only make editable if there are contents
		endif
		
		Make/W/U/O/N=(0,3) WD_logInfoColors = 65535
		WD_logInfoColors[0][]  = {{255},{255},{255}};	SetDimLabel 0, 0, white,		WD_logInfoColors	// define some colors
		WD_logInfoColors[1][]  = {{255},{243},{243}};	SetDimLabel 0, 1, red,			WD_logInfoColors
		WD_logInfoColors[2][]  = {{255},{243},{252}};	SetDimLabel 0, 2, pink,			WD_logInfoColors
		WD_logInfoColors[3][]  = {{252},{243},{255}};	SetDimLabel 0, 3, purple,		WD_logInfoColors
		WD_logInfoColors[4][]  = {{243},{243},{255}};	SetDimLabel 0, 4, blue,			WD_logInfoColors
		WD_logInfoColors[5][]  = {{243},{255},{255}};	SetDimLabel 0, 5, cyan,			WD_logInfoColors
		WD_logInfoColors[6][]  = {{243},{255},{243}};	SetDimLabel 0, 6, green,		WD_logInfoColors
		WD_logInfoColors[7][]  = {{255},{255},{240}};	SetDimLabel 0, 7, yellow,		WD_logInfoColors
		WD_logInfoColors[8][]  = {{255},{243},{233}};	SetDimLabel 0, 8, orange,		WD_logInfoColors
		WD_logInfoColors[9][]  = {{243},{243},{243}};	SetDimLabel 0, 9, gray,			WD_logInfoColors
		WD_logInfoColors[10][] = {{220},{240},{255}};	SetDimLabel 0,10, 'dark blue',	WD_logInfoColors
		WD_logInfoColors[11][] = {{220},{255},{220}};	SetDimLabel 0,11, 'dark green',	WD_logInfoColors
		WD_logInfoColors[12][] = {{255},{220},{220}};	SetDimLabel 0,12, 'dark red',	WD_logInfoColors
		WD_logInfoColors[13][] = {{110},{160},{255}};	SetDimLabel 0,13, selection,	WD_logInfoColors	// the last color is reserved for selections
		WD_logInfoColors *= 257
		
		if (DimSize(list,1) > 1)																// recolor columns after reopening the panel
			SpectraWatchDog#WD_colorListWave(WD_logInfoSelList)
		endif
	SetDataFolder savDF
	return 0
end

// *** UI list hook function *** #################################################################

function WD_ListHook(STRUCT WMListboxAction &s) : ListboxControl
//	if (s.eventCode <3 || s.eventCode > 5)														// debug
//		Print s.eventCode, s.eventMod
//	endif
	
	int retValue = 0
	Switch (s.eventCode)
		case 1:		// mouse-down
			if (s.eventMod & 16)																// right-click menu
				retValue = 1
				if (s.row == -1)																// title row
					DFREF stg = root:Packages:$(kWatchdogWorkingDir)
					SVar colRepeat = stg:WD_RepeatColumn
					SVar targetExt = stg:WD_TargeFileExt
					SVar colorList = stg:WD_ColumnColors
					
					wave/T wRepeat = ListToTextWave(colRepeat, ",")
					ReDimension/N=(max(DimSize(wRepeat,0),DimSize(s.listWave,1))) wRepeat		// make sure the list is not too short
					wRepeat = SelectString(numtype(str2num(wRepeat[p])), wRepeat[p], "0")
					int isRepeated = str2num(wRepeat[s.col])
					
					PopupContextualMenu coloList(s.colorWave)+"Default Colors;-;Default Widths;Optimal Widths;Equal Widths;-;"+SelectString(isRepeated,"","!*")+"Repeat Values;"
					if (V_Flag > 0)
						if (V_Flag-1 < DimSize(s.colorWave,0)-1)
							s.SelWave[][s.col][1] = V_Flag-1									// color column
						endif
						if (StringMatch(S_selection,"Default Colors"))
							colorList = WD_getListDefaults(targetExt, WDInfo_DefaultColumnColors)
							SpectraWatchDog#WD_colorListWave(s.SelWave)
						endif
						if (StringMatch(S_selection,"* Widths"))
							string colWidths = "1"												// default is equal widths
							StrSwitch(S_selection)
								case "Default Widths":
									colWidths = WD_getListDefaults(targetExt, WDInfo_DefaultColumnWidths)
									colWidths = WD_checkColWidths(colWidths)
								break
								case "Optimal Widths":											// tries to find optimal column widths from the contents
									Make/Free/N=(DimSize(s.listWave,0),DimSize(s.listWave,1)) tempLen = strlen(s.listWave[p][q])
									MatrixOp/Free tempMat = maxCols(tempLen)^t
									MatrixOp/Free tempLen = replace(tempMat,0,5)/minVal(replace(tempMat,0,inf))
									tempLen = tempLen[p] < mean(tempLen)/2 ? round(mean(tempLen)/2) : round(tempLen[p])
									wfprintf colWidths, "%d,", tempLen;	colWidths = RemoveEnding(colWidths,",")
								break
							EndSwitch
							Execute/Q/Z "ListBox WD_DataDisplayList ,win="+s.win+" ,widths={"+colWidths+"}"
						endif
						if (StringMatch(S_selection,"Repeat Values"))
							wRepeat[s.col] = num2str(!isRepeated)								// toggle repeat setting
							wfprintf colRepeat, "%s,", wRepeat
							colRepeat = RemoveEnding(colRepeat,",")
						endif
						SpectraWatchDog#WD_saveSettingsToFile()
					endif
				else
					PopupContextualMenu "Undo \ Redo;-;Cut;Copy;Paste;-;Select Whole Column;Select All Cells;-;Delete Selected Entries;"
					if (V_Flag > 0 && s.row < DimSize(s.listWave, 0))
						Switch (V_flag)
							case 1:
								SpectraWatchDog#doUndoRedo()
							break
							case 3:
							case 4:
							case 5:
								SpectraWatchDog#doCutCopyPaste(S_selection)
							break
							case 7:
								s.SelWave[][][0]	  = s.SelWave[p][q] & ~(2^0)				// first deselect all
								s.SelWave[][s.col][0] = s.SelWave[p][q] | (2^0)					// select column
							break
							case 8:
								s.SelWave[][][0]	  = s.SelWave[p][q] | (2^0)					// select all
							break
							case 10:
								s.listWave = SelectString((s.SelWave[p][q][0] & 2^0) != 0 || (s.SelWave[p][q][0] & 2^3) != 0, s.listWave[p][q], "")
								SpectraWatchDog#WD_updateLog()
							break
						EndSwitch
					endif
				endif
			else	// select and color title cell
				if (s.row == -1)																// title row
					s.SelWave[][][0]	  = s.SelWave[p][q] & ~(2^0)							// first deselect all
					s.SelWave[][s.col][0] = s.SelWave[p][q] | (2^0)								// select column
				endif
				SpectraWatchDog#WD_colorListWave(s.SelWave)										// color the tag of all selected rows
				int i,j
				for (i = 0; i < DimSize(s.SelWave,0); i++)
					for (j = 0; j < DimSize(s.SelWave,1); j++)
						if ((s.SelWave[i][j][0] & 2^0) != 0 || (s.SelWave[i][j][0] & 2^3) != 0)
							s.SelWave[i][0][1] = (DimSize(s.colorWave,0)-1)
						endif
					endfor
				endfor
			endif
		break
		case 6:		// begin edit
			SpectraWatchDog#WD_colorListWave(s.SelWave)
			s.SelWave[s.row][0][1] = (DimSize(s.colorWave,0)-1)									// color tag
		break
		case 7:		// end edit
			SpectraWatchDog#WD_updateLog()
		break
		case 11:	// column resize
			SpectraWatchDog#WD_saveSettingsToFile()
		break
	EndSwitch
	return retValue
end

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

static function/S coloList(wave colors)
	int i; string list = "", str = ""
	for (i=0; i<DimSize(colors,0)-1; i++)
		sprintf str, "[%.2d] %s;", i, GetDimLabel(colors,0,i);	list += str
	endfor
	return list
end

// *** UI panel hook function *** ################################################################

function WD_PanelHook(STRUCT WMWinHookStruct &s)
//	if (s.eventCode <3 || s.eventCode > 5)														// debug
//		Print s.eventCode, s.eventName, s.keycode, s.menuname, s.menuitem
//	endif

	DFREF stg = root:Packages:$(kWatchdogWorkingDir)
	wave/T logList = stg:WD_logInfoDispList
	wave   selList = stg:WD_logInfoSelList
	
	variable hookTakeover = 0
	StrSwitch (s.eventName)
		case "menu":
			StrSwitch (s.menuitem)
				case "Copy":
					hookTakeover = 1
				case "Paste":
					SpectraWatchDog#doCutCopyPaste(s.menuitem)
				break
				case "Select All":
					//selList[][][0] = selList[p][q][r] | (2^0)
				break
			EndSwitch
		break
		case "earlyKeyboard":																	// only works in Igor 9
			Switch (s.keycode)
				case 65:	// ctrl + A
					selList[][][0] = selList[p][q][r] | (2^0)
				break
				case 88:	// ctrl + X
					SpectraWatchDog#doCutCopyPaste("Cut")
				break
				case 90:	// ctrl + Z
					SpectraWatchDog#doUndoRedo()
				break
				case 127:	// del key
					logList = SelectString((selList[p][q][0] & 2^0) != 0 || (selList[p][q][0] & 2^3) != 0, logList[p][q], "")
					SpectraWatchDog#WD_updateLog()
				break
				case 28:	// cursor keys
				case 29:
				case 30:
				case 31:
					Execute/Q/Z/P "SpectraWatchDog#WD_cursorColor()"							// needs to be executed after selection completes
				break
			EndSwitch
		case "keyboard":
		break;
		case "kill":
		break
		case "resize":
			ListBox WD_DataDisplayList ,win=$s.winName ,pos={11,85} ,size={s.winRect.right-22,s.winRect.bottom-96}
		break
	EndSwitch
	return hookTakeover
end

//################################################################################################

static function WD_cursorColor()																// dedicated function to color tags after keyboard cursor movement
	DFREF stg = root:Packages:$(kWatchdogWorkingDir)
	wave selList = stg:WD_logInfoSelList
	wave colWave = stg:WD_logInfoColors
	SpectraWatchDog#WD_colorListWave(selList)
	int i, j, select_color = DimSize(colWave,0)-1
	for (i = 0; i < DimSize(selList,0); i++)
		for (j = 0; j < DimSize(selList,1); j++)
			if ((selList[i][j][0] & 2^0) != 0)
				selList[i][0][1] = select_color
			endif
		endfor
	endfor
	return 0
end

//################################################################################################

static function doUndoRedo()
	DFREF stg = root:Packages:$(kWatchdogWorkingDir)
	wave/T   logList = stg:WD_logInfoDispList
	wave/T/Z logBack = stg:WD_logInfoListBackup
	if (!WaveExists(logBack))
		return -1
	endif
	variable minRow = min(DimSize(logList,0),DimSize(logBack,0))-1
	variable minCol = min(DimSize(logList,1),DimSize(logBack,1))-1
	logList[,minRow][,minCol] = logBack[p][q]
	SpectraWatchDog#WD_updateLog()
	return 0
end

//################################################################################################

static function doCutCopyPaste(string which)
	DFREF stg = root:Packages:$(kWatchdogWorkingDir)
	wave/T logList = stg:WD_logInfoDispList
	wave   selList = stg:WD_logInfoSelList
	
	int i, j, counter = 0, delete = 0
	StrSwitch (which)
		case "Cut":
			delete = 1
		case "Copy":
			string CopyStr = ""
			for (i = 0; i < DimSize(selList,0); i++)											// Excel-compatible copy function
				for (j = 0; j < DimSize(selList,1); j++)
					if ((selList[i][j][0] & 2^0) != 0 || (selList[i][j][0] & 2^3) != 0)			// if cell is selected append contents to copy-list
						if (counter == 0)														// find first selected row
							counter = i
						endif
						if(counter < i)															// step to next row has occurred
							CopyStr = RemoveEnding(CopyStr, "\t") + "\r\n"
							counter = i
						endif
						CopyStr += logList[i][j] + "\t"
						if (delete)
							logList[i][j] = ""
						endif
					endif
				endfor
			endfor
			CopyStr = RemoveEnding(CopyStr, "\t") + "\r\n"
			PutScrapText CopyStr
			if (delete)
				SpectraWatchDog#WD_updateLog()
			endif
		break
		case "Paste":
			string PasteStr = GetScrapText()													// paste directly into the list
			PasteStr = ReplaceString("\r", ReplaceString("\n", ReplaceString("\t", PasteStr, ";"), ";"), ";")
			PasteStr = RemoveFromList("", PasteStr)
			for (i = 0; i < DimSize(selList,0); i++)
				for (j = 0; j < DimSize(selList,1); j++)
					if ((selList[i][j][0] & 2^0) != 0 || (selList[i][j][0] & 2^3) != 0)
						logList[i][j] = StringFromList(counter,PasteStr)						// successively paste contents into selected cells
						counter += counter == ItemsInList(PasteStr)-1 ? 0 : 1					// if the end of the list is reached, repeat last element
					endif
				endfor
			endfor
			SpectraWatchDog#WD_updateLog()
		break
	EndSwitch
end

//################################################################################################

static function WD_colorListWave(wave selList)													// applies color to certain entries
	DFREF stg = root:Packages:$(kWatchdogWorkingDir)
	SVar colorList = stg:WD_ColumnColors
	wave colorWave = stg:WD_logInfoColors
	variable i, color, colorNum = ItemsInList(colorList,",")
	for (i = 0; i < DimSize(selList,1); i++)
		color = str2num(StringFromList(i,colorList,","))
		color = numtype(color)!= 0 || i>=colorNum ? 0 : color
		color = color < DimSize(colorWave,0)-1 ? color : 0										// make to not exceed usable colors
		selList[][i][1] = color
	endfor
	return 0
end

// *** UI settings handling *** ##################################################################

static function WD_saveSettingsToFile([int reset])
	DFREF stg = root:Packages:$(kWatchdogWorkingDir)
	SVAR TargetDF  = stg:WD_TargetFolder
	SVAR logFPath  = stg:WD_LogFilePath
	wave/T logList = stg:WD_logInfoDispList
	wave   selList = stg:WD_logInfoSelList
	wave colorWave = stg:WD_logInfoColors
	SVar targetExt = stg:WD_TargeFileExt
	SVar colRepeat = stg:WD_RepeatColumn
	SVar colorList = stg:WD_ColumnColors
	SVar colWidths = stg:WD_ColumnWidths	
	
	int doReset = ParamIsDefault(reset) ? 0 : reset
	int special = DimSize(colorWave,0)-1
	if (doReset)
		colorList = WD_getListDefaults(targetExt, WDInfo_DefaultColumnColors)
		colRepeat = WD_getListDefaults(targetExt, WDInfo_DefaultRepeatColumn)
		colWidths = WD_getListDefaults(targetExt, WDInfo_DefaultColumnWidths)
		colWidths = WD_checkColWidths(colWidths)
		SpectraWatchDog#WD_colorListWave(selList)
		Execute/Q/Z "ListBox WD_DataDisplayList ,win=SpectraWatchDogPanel ,widths={"+colWidths+"}"
	elseif (DimSize(logList,1) > 1)
		Duplicate/Free/RMD=[0][][1] selList, tempColors											// save the current color settings if there are already some entries
		if (tempColors[0] >= special)															// this color is reserved for highlighting
			Duplicate/Free/RMD=[][0][1] selList, tempFirstColor
			Extract/FREE/INDX tempFirstColor, colorIndex, tempFirstColor<special				// see if one of the rows is colored normally
			if (DimSize(colorIndex,0) > 0)
				tempColors[0] = tempFirstColor[colorIndex[0]]
			else
				tempColors[0] = str2num(StringFromList(0,colorList,","))						// don't save the color for now
			endif
		endif
		wfprintf colorList, "%d,", tempColors
		colorList = RemoveEnding(colorList,",")
		ControlInfo/W=SpectraWatchDogPanel WD_DataDisplayList									// save the current width settings
		colWidths = S_columnWidths
	endif
	
	if (!kUseSettingsFile)
		return 0
	endif
	
	string Savestring = ""
	Savestring += "Target=" + ParseFilePath(5, TargetDF, "*", 0, 0) + "\r\n"
	Savestring += "LogDir=" + ParseFilePath(5, logFPath, "*", 0, 0) + "\r\n"
	Savestring += "Spectra_Extension=" + targetExt + "\r\n"
	Savestring += "Column_Widths=" + colWidths + "\r\n"
	Savestring += "Column_Repeat=" + colRepeat + "\r\n"
	Savestring += "Column_Colors=" + colorList + "\r\n"
	
	int fileID = 0
	Open/Z fileID as ParseFilePath(1, FunctionPath(""), ":", 1, 0)+kSettingsFileName
	if (fileID != 0)
		FPrintF fileID, SaveString																// write settings => has to match with READING CODE in WD_prepareWorkingFolder() etc.
		Close fileID
	endif
end

//################################################################################################

static function/S WD_readSettingsFromFile()
	int FileID
	string read = "", output = ""	
	Open/Z/R fileID as ParseFilePath(1, FunctionPath(""), ":", 1, 0)+kSettingsFileName
	if (!V_flag)
		do
			FReadLine fileID, read
			output += read
		while (strlen(read) > 0)
		Close fileID
	endif
	return ReplaceString("\r", output, ";")														// FReadLine terminates just with a line break
end

//################################################################################################

function WD_SelectInputFolderProc(STRUCT WMButtonAction &s) : ButtonControl
	if (s.eventCode != 2)
		return 0
	endif
	
	PathInfo WD_WatchFolderPath																	// check path status
	if (!V_flag)
		NewPath/O/Q WD_WatchFolderPath, SpecialDirPath("Desktop", 0, 0, 0)						// preset folder to the Desktop
	endif
	PathInfo/S WD_WatchFolderPath																// set open path to Desktop or last folder
	NewPath/O/M="Select folder for the watchdog process to watch."/Q WD_WatchFolderPath			// open folder dialog
	if (V_Flag != 0)
		return -1
	endif
	PathInfo WD_WatchFolderPath																	// keep path in global variable as well
	DFREF stg = root:Packages:$(kWatchdogWorkingDir)
	SVar TargetDir = stg:WD_TargetFolder
	TargetDir = ParseFilePath(5, S_path, "*", 0, 0)
	
	SpectraWatchDog#WD_saveSettingsToFile()
	return 0
end

//################################################################################################

function WD_SelectLogFileFolderProc(STRUCT WMButtonAction &s) : ButtonControl
	if (s.eventCode != 2)
		return 0
	endif
	DFREF stg = root:Packages:$(kWatchdogWorkingDir)
	SVAR logFilePath = stg:WD_LogFilePath
	int fileID
	Open/D/F="Log Files (*"+kLogFileEnding+"):"+kLogFileEnding+";"/M="Chose the log name and location." fileID as logFilePath
	if(strlen(S_fileName))
		logFilePath = ParseFilePath(5, S_fileName, "*", 0, 0)									// the log file path only exists in the global and not as PATH variable
		SpectraWatchDog#WD_saveSettingsToFile()
		SpectraWatchDog#WD_askForLoadLog()
	endif
	return 0
end

//################################################################################################

function WD_SetLogFileFolderProc(STRUCT WMSetVariableAction &s) : SetVariableControl
	if (s.eventCode == 1 || s.eventCode == 2)
		int resetList = 0
		StrSwitch(s.ctrlName)
			case "v_LogFileFolder":
				SpectraWatchDog#WD_askForLoadLog()
			break
			case "v_TargetFileExt":
				if (ItemsInList(getFuncList(s.sval, "WDfunc_", 8, 8192)))
					DoAlert/T="Reload Default Settings?" 1, "Do you want to load the default table settings for "+UpperStr(s.sval)+" files?"
					resetList = (V_Flag == 1)
				endif
			break
		EndSwitch
		SpectraWatchDog#WD_saveSettingsToFile(reset=resetList)
	endif
	return 0
end

//################################################################################################

static function WD_askForLoadLog()
	DFREF stg = root:Packages:$(kWatchdogWorkingDir)
	wave/T logList = stg:WD_logInfoDispList
	GetFileFolderInfo/Z/Q getLogPath()															// if the log file already exists
	if (V_flag == 0 && V_isFile == 1)
		DoAlert/T="Log File Already Exists" 1, "Selected log file already exists. Load previous log contents into current session?"
		if (V_Flag == 1)
			SpectraWatchDog#WD_ReadFromLog()
		endif
	else
		if (strlen(logList[0][0]))
			DoAlert/T="Initialize New Log File" 1, "A new log file will be created. Also clear entries from currently displayed list (saved log entries will NOT be deleted)?"
			if (V_Flag == 1)
				SpectraWatchDog#WD_ClearAndReset("ClearLogList")
			endif
		endif
	endif
	return 0
end

// *** UI control helper functions *** ############################################################

function WD_StartTaskButton(STRUCT WMButtonAction &s) : ButtonControl
	if (s.eventCode != 2)
		return 0
	endif
	
	DFREF stg = root:Packages:$(kWatchdogWorkingDir)
	SVAR targetDir	= stg:WD_TargetFolder
	SVAR targetExt	= stg:WD_TargeFileExt
	SVAR logFilePath= stg:WD_LogFilePath
	SVAR doneList	= stg:WD_ProcessedFileList
	NVAR startCount	= stg:WD_startCount
	NVAR runCount	= stg:WD_runCounter
	
	CtrlNamedBackground SpectraWatchDogTask, status
	if (NumberByKey("RUN", S_info))	// running
		CtrlNamedBackground SpectraWatchDogTask, stop
	else							// stopped
		string pathSep = SelectString(strsearch(targetDir,"\\",0) > -1,":","\\")
		targetDir = ParseFilePath(2, targetDir, pathSep, 0, 0)									// add ending separator, if not already there
		GetFileFolderInfo/Z/Q targetDir
		if (V_flag != 0 || V_isFolder != 1)
			DoAlert/T="Error: Folder not found." 0, "The selected target folder does not seem to be valid. Set a valid folder to watch first."
			return -1
		endif
		GetFileFolderInfo/Z/Q ParseFilePath(1, ParseFilePath(5, logFilePath, "\\", 0, 0), "\\", 1, 0)
		if (V_flag != 0 || V_isFolder != 1)
			DoAlert/T="Error: Folder not found." 0, "The folder for the log file does not seem to be valid. Set a valid folder to write the log file to."
			return -2
		endif
		
		NewPath/O/Q WD_WatchFolderPath, targetDir												// make sure the path is properly set (even if entered manually)
		ControlInfo/W=$s.win c_ReadOnlyNew														// if already present files should be ignored
		if (V_Value)
			int i; string currList = ""
			for (i = 0; i < ItemsInList(targetExt); i++)
				currList += IndexedFile(WD_WatchFolderPath,-1,"."+StringFromList(i,targetExt))	// read all files in the current folder into processed list to check them off
			endfor
			doneList = RemoveFromList(currList, doneList) + currList
		endif
		
		GetFileFolderInfo/Z/Q logFilePath														// if the log file already exists
		if (V_flag == 0 && V_isFile == 1)
			wave/T/Z logFileData = SpectraWatchDog#WD_loadCurrentLog()
			if (WaveExists(logFileData))
				string logFileNames = processLogNameList(logFileData, extractZip=1)
				doneList = RemoveFromList(logFileNames, doneList) + logFileNames
			endif
		endif
		
		if (CmpStr(s.ctrlName,"b_ReadOnceButton") == 0)											// just briefly check folder and don't start the task
			SpectraWatchDog#WD_checkNewFiles()
		else
			startCount = ticks; runCount = 0
			CtrlNamedBackground SpectraWatchDogTask, period=kCheckFolderPeriod, proc=WD_SpectraWatchFolderBT, dialogsOK=0
			CtrlNamedBackground SpectraWatchDogTask, start
		endif
	endif
	
	SpectraWatchDog#enableDisableControls()
	return 0
end

//################################################################################################

function WD_ReadFromLogButton(STRUCT WMButtonAction &s) : ButtonControl
	if (s.eventCode == 2)
		SpectraWatchDog#WD_ReadFromLog()
	endif
	return 0
end

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

static function WD_ReadFromLog()
	DFREF stg = root:Packages:$(kWatchdogWorkingDir)
	SVAR logFNames = stg:WD_logInfoFileNames
	SVAR doneFList = stg:WD_ProcessedFileList
	wave/T logList = stg:WD_logInfoDispList
	wave   selList = stg:WD_logInfoSelList
	
	wave/Z/T logFileData = SpectraWatchDog#WD_loadCurrentLog()
	if (!WaveExists(logFileData))
		DoAlert 0, "Log file does not exist (yet)."
		return -1
	endif
	
	string zipFNames = ""
	logFNames = processLogNameList(logFileData, extractZip=0)
	zipFNames = processLogNameList(logFileData, extractZip=1)
	doneFList = RemoveFromList(zipFNames, doneFList) + zipFNames
	
		Redimension/N=(DimSize(logFileData,0)-1,-1,-1) logList, selList							// update display list
	if (DimSize(logList,1) < DimSize(logFileData,1)-1)
		Redimension/N=(-1,DimSize(logFileData,1)-1,-1) logList, selList
	endif
	int i
	for (i = 0; i < min(DimSize(logFileData,1)-1,DimSize(logList,1)); i++)						// write labels from the table
		string title = cleanStr(logFileData[0][i+1],"\"|'|,|;|")[0,255]
		SetDimLabel 1,i,$title, logList
	endfor
	logList = logFileData[p+1][q+1]
	selList[][][0] = 0x02
	SpectraWatchDog#WD_colorListWave(selList)
	return 0
end

//################################################################################################

function WD_ClearAndResetButton(STRUCT WMButtonAction &s) : ButtonControl
	if (s.eventCode == 2)
		SpectraWatchDog#WD_ClearAndReset(ReplaceString("b_",s.ctrlName,""))
	endif
	return 0
end

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

static function WD_ClearAndReset(string which)
	DFREF stg = root:Packages:$(kWatchdogWorkingDir)
	SVAR doneFList = stg:WD_ProcessedFileList
	SVAR logFNames = stg:WD_logInfoFileNames
	SVAR logFPath  = stg:WD_LogFilePath
	wave/T logList = stg:WD_logInfoDispList
	wave   selList = stg:WD_logInfoSelList
	NVAR timeCount = stg:WD_runCounter
	StrSwitch (which)
		case "ResetWatchDog":
			DoAlert/T="Reset to initial state" 2, "Resetting watchdog to initial state ... (Press Cancel to abort)\r\rDo this to reprocess files into a new log. Already saved log files are NOT deleted or altered.\r\rAlso set log file name to a new file?\r(Press No to leave log name unchanged.)"
			if (V_Flag == 3)
				break
			endif
			if (V_Flag == 1 && StringMatch(logFPath,"*"+kLogFileEnding))						// increment log file name
				string pathSep = SelectString(strsearch(logFPath,"\\",0)>-1,":","\\")
				string loadDF  = ParseFilePath(1, logFPath, pathSep, 1, 0)
				string logFile = ParseFilePath(3, logFPath, pathSep, 1, 0), num
				SplitString/E="(\d+)(?!.*\d)" logFile, num
				logFile = RemoveEnding(logFile,num)
				int i, doSave = 0
				for (i=1; i<100; i++)
					string newFile = loadDF + logFile + num2str(i) + kLogFileEnding
					GetFileFolderInfo/Z/Q newFile
					if (V_flag)
						logFPath = newFile
						doSave = 1
						break
					endif
				endfor
				if (doSave)
					SpectraWatchDog#WD_saveSettingsToFile()
				endif
			endif
			doneFList = ""
			timeCount = 0
		case "ClearLogList":
			Redimension/N=(1,1,-1) logList,selList
			logFNames = ""																		// make sure to clear both logNames and logList
			logList   = ""
			selList   = 0
		break
	EndSwitch
	return 0
end

//################################################################################################

static function/S processLogNameList(wave/T logData [,int extractZip])
	int i, doZip = ParamIsDefault(extractZip) ? 0 : extractZip									// extract only the name of zip files if present => zip contents are not relevant for file watcher
	Make/Free/T/N=(DimSize(logData,0)-1) fileList = logData[p+1][0]								// save loaded file names into processed and logged list
	Extract/FREE/INDX fileList, emptyList, !strlen(fileList)									// delete empty entries
	for (i=DimSize(emptyList,0)-1; i>-1; i--)
		DeletePoints/M=0 emptyList[i],  1, fileList
		DeletePoints/M=0 emptyList[i]+1,1, logData
	endfor
	if (doZip)
		string zipFileName = ""
		for (i=DimSize(fileList,0)-1; i>-1; i--)
			SplitString/E=("^\[(.*)\]") fileList[i], zipFileName								// extract zip file name
			fileList[i] = SelectString(strlen(zipFileName),fileList[i],zipFileName)
		endfor
	endif
	FindDuplicates/Free/RT=fileList_clean fileList												// remove duplicate entries
	string logListNames = ""; wfprintf logListNames, "%s;", fileList_clean
	return logListNames
end

// *** live log updater *** ######################################################################

static function WD_updateLog()
	DFREF stg = root:Packages:$(kWatchdogWorkingDir)
	SVAR logFPath = stg:WD_LogFilePath
	SVAR logNames = stg:WD_logInfoFileNames
	wave/T logList= stg:WD_logInfoDispList
	
	wave/T/Z logFileData = SpectraWatchDog#WD_loadCurrentLog()
	if (!WaveExists(logFileData))
		return -1
	endif
	if (!SpectraWatchDog#FileIsNotLocked(logFPath,1))											// file successfully opened?
		DoAlert 0, "The log file is locked by another program and cannot be updated. Make sure the log file is not open at the same time, e.g., in Excel."
		return -2
	endif
	if (ItemsInList(logNames) != DimSize(logList,0))
		DoAlert 0, "Somehow the local log table is broken and the log file cannot be updated. Try to re-read the log entries from the file to fix this."
		return -3
	endif
	
	int rows = DimSize(logFileData,0), i
	int cols = DimSize(logFileData,1)
	
	if (cols != DimSize(logList,1)+1)															// make sure the local and loaded lists have the same no. of columns
		DoAlert 0, "The log file has the wrong number of entries and cannot be updated. Check the file for additional or missing columns."
		return -4
	endif
	
	Make/FREE/T/N=(rows)		  tempLogName = logFileData[p][0]
	Make/FREE/T/N=(rows-1,cols-1) tempLogData = logFileData[p+1][q+1]							// for backup
	string namesInFile = "";	wfprintf namesInFile, "%s;", tempLogName						// a list of file names in the log file including possible invalid ones
	
	for (i = 0; i < ItemsInList(logNames); i++)
		int entry = WhichListItem(StringFromList(i, logNames), namesInFile)						// find entry in the log and replace the contents
		if (entry > 0)
			logFileData[entry][1,] = logList[i][q-1]
		endif
	endfor
	
	// to make the log compatible with excel for entries which contain the delimiter character (e.g., comma), add "" around such strings
	Make/Free/D/N=(rows, cols) findDelim = strsearch(logFileData[p][q],kDelimiterChar,0)
	logFileData = SelectString((findDelim[p][q] > -1), logFileData[p][q], "\""+logFileData[p][q]+"\"")
	
	Save/O/A=0/J/M="\r\n"/DLIM=kDelimiterChar logFileData as logFPath							// overwrite log file
	if (!EqualWaves(tempLogData, logList, 1))
		Duplicate/O tempLogData, stg:WD_logInfoListBackup										// backup for undo functionality
	endif
	return 0
end

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

static function/WAVE WD_loadCurrentLog()
	GetFileFolderInfo/Z/Q getLogPath()
	if (V_flag != 0 || V_isFile != 1)
		return $""
	endif
	DFREF saveDF = GetDataFolderDFR()
	SetDataFolder NewFreeDataFolder()
		LoadWave/J/O/K=2/A/V={kDelimiterChar, "", 0, 0}/Q getLogPath();	int success = V_flag	// load log file
		if (success)
			Concatenate/Free/T WaveList("wave*",";","TEXT:1"), loadedData						// create one big table
		endif
	SetDataFolder saveDF
	if (success)
		return loadedData
	else
		return $""
	endif
end